// MemBrain C Code export library V06.00.00.00

/*
--- License information and liability disclaimer ---
------------------
The source code exported by MemBrain including all supplementary source code
provided as part of the MemBrain installation is subject to licensing according
to the terms indicated in the license agreement that is displayed when
installing the MemBrain software.

You can also download and view this license agreement on the MemBrain
homepage (http://www.membrain-nn.de).

The terms of the license agreement do especially prohibit the commercial use of
any of the MemBrain source code items for a commercial purpose except the
user or company that intends to use the code in a commercial way is a
commercially registered MemBrain user and the owner of a corresponding commercial
MemBrain license key.

The MemBrain program and all associated supplementary software including but not
limited to the MemBrain DLL and all MemBrain generated source code are distributed
"as is". No warranty of any kind is expressed or implied. You use it at your own risk.
Thomas Jetter will not be liable for data loss, damages and loss of profits or
properties or any other kind of damages or loss caused by use or misuse of this
software.

Copyright 2003 - 2019 Thomas Jetter
*/


#ifndef _NEURON_H_
#define _NEURON_H_

// possible input functions
enum
{
    MB_IF_SUM,
    MB_IF_MUL,
    MB_IF_MAX,
    MB_IF_AVG,
    MB_IF_RAND
};

/// Possible activation functions
enum
{
	MB_AF_LOGISTIC,
	MB_AF_IDENTICAL,
	MB_AF_IDENTICAL_0_1,
	MB_AF_TAN_H,
	MB_AF_BINARY,
    MB_AF_MIN_EUCLID_DIST,
    MB_AF_IDENTICAL_M1_1,
    MB_AF_RELU,
    MB_AF_SOFTPLUS,
    MB_AF_BIN_DIFF,
    MB_AF_SOFTMAX
};


/// Possible output fire levels
enum
{
	MB_OFL_1,		        ///< 1
	MB_OFL_ACT  		    ///< Activation
};


/// Neuron parameters common to all types of neurons
typedef struct
{
#ifdef NEED_INPUT_FUNC
  UINT8           inputFunc;            ///< Input function
#endif
  UINT8           actFunc;              ///< Activation function

#ifdef NEED_OUT_BUFFER
  T_OUT_IDX       outIdx;               ///< Index of first output value in output buffer
#endif

#ifdef OUT_FIRE_LEVEL_NOT_ACT
  UINT8           fireLevel;            ///< Output Fire Level
#endif

#ifdef SUSTAIN_FACT
  MB_FLOAT         sustainFact;          ///< Activation Sustain Factor
#endif

#ifdef FIRE_DELAY
  T_FIRE_DELAY    recovTime;            ///< Output Recovery Time
#endif

#ifdef LINK_LENGTHS
  T_LINK_LEN      maxOutLinkLen;        ///< Length of longest output link
#endif

#ifdef FIRE_THRESHOLDS
  MB_FLOAT         fireThresLow;         ///< Lower Fire Threshold
  MB_FLOAT         fireThresHi;          ///< Upper Fire Threshold
#endif

} SNeuronParms;




/// Additional parameters of input neurons
#ifdef INPUT_NORMALIZATION
typedef struct
{

  BOOL            useNormalize;         ///< Use normalization?
  T_NORM_FLOAT    normalizeLow;         ///< Lower normalization limit
  T_NORM_FLOAT    normalizeHigh;        ///< Upper normalization limit

} SInputNeuronParms;
#endif



/// Additional parameters of hidden neurons
typedef struct
{

  MB_FLOAT         actThres;             ///< Activation Threshold
  T_LINK_IDX      inLinkIdx;            ///< Index of first input link
  T_IN_LINK_COUNT inLinkCount;          ///< Number of input links

#ifdef ACT_FUNC_PARAMETRIZED
  MB_FLOAT         actFuncParm;          ///< Activation function parametrization
#endif

#ifdef LEAKAGE_PARAMETRIZED
  MB_FLOAT         leakageParm;          ///< Activation function parametrization
#endif

} SHiddenNeuronParms;




/// Additional parameters of output neurons
typedef struct
{
  MB_FLOAT        actThres;             ///< Activation Threshold
  T_LINK_IDX      inLinkIdx;            ///< Index of first input link
  T_IN_LINK_COUNT inLinkCount;          ///< Number of input links

#ifdef ACT_FUNC_PARAMETRIZED
  MB_FLOAT         actFuncParm;          ///< Activation function parametrization
#endif

#ifdef LEAKAGE_PARAMETRIZED
  MB_FLOAT         leakageParm;          ///< Activation function parametrization
#endif

#ifdef OUTPUT_NORMALIZATION
  BOOL            useNormalize;         ///< Use normalization?
  T_NORM_FLOAT    normalizeLow;         ///< Lower normalization limit
  T_NORM_FLOAT    normalizeHigh;        ///< Upper normalization limit
#endif

} SOutputNeuronParms;



/// Data common to all types of neurons (besides buffered output value(s))
typedef struct
{
  T_ACT_FLOAT     activation;           ///< The Activation

#ifdef LINK_LENGTHS
  T_OUT_IDX       outIdxCurrent;        ///< Current index into output buffer array
#endif

#ifdef FIRE_DELAY
  T_FIRE_DELAY    fireDelay;            ///< Current fire delay countdown value
#endif

#ifdef SUSTAIN_FACT
  BOOL            firstThinkStep;       ///< First think step after applying activation?
#endif

} SNeuronData;

/// Definition for a neuron layer
typedef struct
{
    T_NEURON_COUNT startIdx;
    T_NEURON_COUNT endIdx;
} SLayerDef;


#ifdef __cplusplus
extern "C"
{
#endif


/// Reset the given neuron
void NeuronReset(T_NEURON_COUNT index);

/// Cause a neuron to propagate
void NeuronPropagate(T_NEURON_COUNT index);

/// Cause a neuron to fire (calculate its output)
void NeuronFire(T_NEURON_COUNT index);

/// Get the latest output signal of a neuron
MB_FLOAT NeuronGetLatestOutput(T_NEURON_COUNT index);

#ifdef LINK_LENGTHS
/// Get a past output signal of a neuron
MB_FLOAT NeuronGetPastOutput(T_NEURON_COUNT index, T_LINK_LEN past);
#endif

/// Set the activation of a neuron. <index> can range from 0 to MB_NEURON_COUNT - 1
void NeuronSetActivation(T_NEURON_COUNT index, T_ACT_FLOAT act);

/// Get the activation value of a neuron. <index> can range from 0 to MB_NEURON_COUNT - 1
T_ACT_FLOAT NeuronGetActivation(T_NEURON_COUNT index);

#ifdef MB_ACT_SUM_POST_PROC_LAYER_COUNT
void NeuronPostProcess(T_NEURON_COUNT index, T_ACT_FLOAT layerActSum);
#endif


#ifdef __cplusplus
}
#endif


#endif
